/* 
  Header File for program "atomic.c"
  written by Nancy S. Brickhouse 
  modified by Randall K. Smith
  modified by Ronnie Hoogerwerf
  modified by Randall K. Smith OCT-3-2002
  modified by Ronnie Hoogerwerf OCT-9-2002
  modified by Randall K. Smith JUN-15-2005 
*/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <fitsio.h>
#include "fitsdebug.h"

#define VERSION "2.0"

#define BLOWUP 1.e20

#define MAX_CHI 200.  /* terms below exp(-200) are simply set to zero. */

#define MAXLINELENGTH 132
#ifndef MAXSTRLEN
#define MAXSTRLEN 128
#endif
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define MAXIMUM_Z 99

#ifndef PI
#define PI 3.1415926535
#endif
#ifndef M_E
#define M_E 2.7182818284590452354
#endif
#define UPSILON_COLL_COEFF 8.629e-6  /* = sqrt{2\pi / kB} hbar^2/m_e^{1.5} */
#define BOLTZ 11605.
#define KBOLTZ 8.617385e-8  /* kB = [keV/K] */
#define RYDBERG 0.013605804
#define ERG_KEV 1.60219e-9
#define HC_IN_KEV_A 12.400
#define HC_IN_ERG_A 1.9848e-8   /* = hc (erg A) = 12400 * 1.602e-12 */
#define MIN_EPSILON 5.e-23  /* in phot cm^3 / s */

/* Ionization balance info */
#define COLUMN_ION_STATE "ION_STATE"
#define COLUMN_LOGT "LOGT"
#define COLUMN_ION_POP "ION_POP"
#define MIN_ION_POP 1.e-10


/* Bremsstrahlung */
#define MIN_RELATIVISTIC_T 5.93e7   /* 1% of 511 keV / 8.617e-8 = 5.93e9 K */
#define HUMMER_TYPE 11   /* The file type in the filemap file */
#define GAUNT_FF_TYPE 13 /* The file type in the filemap file */
#define GAUNT_HUMMER_IVAL "I_VALUE"
#define GAUNT_HUMMER_JVAL "J_VALUE"
#define GAUNT_HUMMER_DVAL "COEFFICIENT"
#define GAUNT_FF_U  "LOG10_U"
#define GAUNT_FF_Z  "Z"
#define GAUNT_FF_NG "N_GAMMA2"
#define GAUNT_FF_G2 "LOG10_GAMMA2"
#define GAUNT_FF_GF "GAUNT_FF"
#define HUMMER 1
#define KELLOGG 2
#define RELATIVISTIC 3
#define BREMS_NONE 4
 
/* Binning types */
#define BIN_LIN_WAVE 1
#define BIN_LOG_WAVE 2
#define BIN_LIN_ENERGY 3
#define BIN_LOG_ENERGY 4
#define BIN_ARBITRARY 5

#define LINEAR 1
#define LOG 2
#define ENTIRE_ION -1

/* Total Ionization/Recombination Rate codes */
#define CI_YOUNGER         101
#define EA_ARNROTH_LITHIUM 101
#define EA_ARNROTH         102
#define EA_MAZZOTTA_IRON   103
#define RR_SHULL  101
#define RR_VERNER 102
#define RR_ARNRAY 103
#define DR_MAZZOTTA 301

/* Radiative recombination numbers */
#define NO_PHOT -1
#define HYDROGENIC 0
#define CLARK 1
#define VERNER 2

/* Satellite line types */
#define ROMANIK 1
#define SAFRONOVA 2

/* General Line types */
#define BOUNDBOUND_LINE 0
#define SATELLITE_LINE  1
#define IONEXC_LINE     2
#define EMISSION_FEATURE_LINE 3

/* Continuum errors */
#define ERROR_RRC    0.01
#define ERROR_TWOPH  0.01
#define ERROR_BREMS  0.01
#define ERROR_PSEUDO 0.01
#define ERROR_FINAL  0.01

#define MIN_XSPEC_EMIS 1.e-38   /* in photons cm^3 s^-1 bin^-1 */

/* Array Limits */
#define MAX_SIG 20 /* Maximum number of parameters for a sigma fit. */
#define MAX_UPS 20 /* Maximum # of paramters for a coll strength fit. */
#define MAX_EMIS 20 /* Maximum # of paramters for a emission feature fit. */
#define DR_PAR 10 /* Max # of parameters for a DR satellite line fit. */
#define MAX_IONREC 20 /* Max # of parameters for a ioniz/recomb fit */

struct ION_INFORMATION{
  char Ion[16];      /* Abbreviated element name */
  int Z;           /* Z for element; total number of electrons in atom */
  int rmJ;         /* charge: rmJ = 1 is neutral atom, etc */
  int nionrec;     /* number of ionization/recombination transitions */
  int nlev;        /* number of levels for model ion */
  long nwv;        /* number of radiative transitions calculated */
  int nexc;        /* number of collisional transitions calculated */
  int npro;        /* number of proton rates calculated */
  int nsatel;      /* number of satellite lines available */
  int nemiss;      /* number of emission features available */
  unsigned long cs_ionrec;/* Data checksum of ion/recomb data */
  unsigned long cs_elvls; /* Data checksum of energy level data */
  unsigned long cs_lines; /* Data checksum of emission line data */
  unsigned long cs_ecoll; /* Data checksum of electron collision data */
  unsigned long cs_pcoll; /* Data checksum of proton collision data */
  unsigned long cs_drsat; /* Data checksum of DR satellite line data */
  unsigned long cs_emiss; /* Data checksum of emission feature data */
};

struct IONREC_LINE_DATA {
  double lambda;     /* wavelength, Angstroms */
  double lambda_err; /* Error in wavelength, Angstroms */
  double br_ratio;   /* unitless branching ratio */
  struct IONREC_LINE_DATA *next; /* linked list */
};

struct IONREC_DATA{
  int ion_final;    /* ionization state (neutral=1) for final ion state */
  int level_final;  /* level of final ion populated by this transition */
  int level_init;   /* level from initial ion state */
  int tr_type;      /* transition type; see datatypes.h for a conversion */
  int tr_index;     /* transition index for radiative transition list */
  int par_type;     /* ioniz/recomb parameterization type */
  double min_temp;  /* minimum temperature, in Kelvin */
  double max_temp;  /* maximum temperature, in Kelvin */
  double Temperature[MAX_IONREC]; /* temperature, in Kelvin */
  double IonRec_Par[MAX_IONREC];  /* ionization/recombination parameters */
  struct IONREC_LINE_DATA *lines;  
                    /* pointer to linked-list of ion/recomb line data */
};

struct LEVEL_DATA{
  int lev;      /* id for level- redundant to store this, but ?useful */
  char config[MAXSTRLEN];  /* Electronic configuration string */
  int deg;      /* statistical weight, or degeneracy = 2 J + 1 */
  int N;        /* principle quantum number */
  int L;        /* orbital quantum number */
  float S;      /* spin quantum number */
  double energy;  /* energy above ground state, eV */
  int phottype; /* Type of photoionization coefficients (default = ?) */
  double sig[MAX_SIG]; /* photoionization from this excited state */
};

struct LINERAD_DATA{
  int lev_upper; /* id for upper level of transition */
  int lev_lower; /* id for lower level of transition */
  double lambda; /* wavelength, Angstroms */
  double lambda_err; /* Error in wavelength, Angstroms */
  /* double lambda_obs; */ /* Observed wavelength, Angstroms */
  /* Removed due to change so that the only lambda used is the observed */
  double A;      /* Einstein transition probability */
};

struct COLLTRNS_DATA{
  int lev_upper; /* id for upper level of transition */
  int lev_lower; /* id for lower level of transition */
  int coll_type; /* Type of coll. strength coeff; Default = Burgess & Tully */
  double min_T; /* minimum temperature for fit */
  double max_T; /* maximum temperature for fit */
  double T[MAX_UPS]; /* Temperature range in interpolation */
  double om[MAX_UPS]; /* fits to collision strength formula (or interp) */   
};

struct SATELLITE_DATA {
  int type; /* Data type */
  int lev_upper;
  int lev_lower;
  double lambda;  /* Angstrom */ 
  double lambda_err;  /* Error in wavelength, Angstroms */ 
  /* double lambda_obs; */ /* Angstrom */ 
  /* Removed due to change so that the only lambda used is the observed */
  double q_exc; /* K^3/2 cm^3 s^-1 */  
  double E_excite;  /* keV */
  double params[DR_PAR]; /* other fit parameters, depending on type of data */
};

struct EMISSION_DATA{
  int type; /* Data type */
  double lambda; /* Angstrom */
  double lambda_err;  /* Error in wavelength, Angstroms */ 
  int lev_upper; /* id for upper level of transition */
  int lev_lower; /* id for lower level of transition */
  int scale_Z;     /* Element to scale emission to */
  int scale_rmJ;   /* Ion to scale emission to */
  int scale_upper; /* Upper level to scale emission to */
  int scale_lower; /* Lower level to scale emission to */
  double min_T; /* minimum temperature for inclusion */
  double max_T; /* maximum temperature for inclusion */
  double T[MAX_EMIS]; /* Temperature range in interpolation */
  double param[MAX_EMIS]; /* Values for emission or fit parameters for it */
};

struct ELEMENT_ION_STATE{
  struct ION_INFORMATION *ion_inf;
  struct IONREC_DATA *ionrec_dat;
  struct LEVEL_DATA *lev_dat;
  struct LINERAD_DATA *linerad_dat;
  struct COLLTRNS_DATA *colltrns_dat;
  struct COLLTRNS_DATA *protrns_dat;
  struct SATELLITE_DATA *satellite_dat;
  struct EMISSION_DATA *emission_dat;
};

typedef struct ELEMENT_ION_STATE * ElementP;

/* ******************************************* */

struct TRANS_RATES{   /* structure to collect all the transition rate data */
  int Z;
  int rmJ;
  int nlev;
  int maxNlev;
  double Te;
  double Tp;
  double *qexc;     /* collisional excitation/deexcitation 
					   and radiative rates */
  double *Arad;
  double *qp;
};

struct LEVEL_RATES {
  int nlev;
  double ionc;            /* ratio of next ion pop to current*/
  double qrecomb_total;   /* total radiative recombination rate */
  double qsatel_total;   /* total dielectronic recombination rate */
  double qioniz_total;   /* total ionization rate */
  double *qrecomb; /* radiative recombination rate to level */
  double *qsatel; /* dielectronic recombination rate to level */
  double *qioniz;  /* ionization rate from level; not currently using this */
};

struct EMISSION_LINES {
  double lambda;
  double lambda_err;
  double epsilon;
  double Plambda;
  int Z;
  int rmJ;
  int ul;
  int ll;
  int type;   /* = 0 for bound-bound; = 1 for DR; = 2 for IE */
  struct EMISSION_LINES *next;
};

struct TWOPHOT_DATA {
  int exists;     /* TRUE if any two phot data exists for this ion. */
  int Z;
  int rmJ;
  double A;       /* in s^-1, the base transition probability */
  double lambda;  /* in A, the base wavelength */
  double lev_pop; /* unitless, the level population relative to ne nH */
};

struct COMPRESS_CONT { /* Holds one ion's continuum only */
  int     Z;
  int     rmJ;
  int     Nbrems;
  double *E_brems; /* in keV */
  double *brems;   /* in photons cm^3 / s / keV */
  int     Ntwoph;
  double *E_twoph; /* in keV */
  double *twoph;   /* in photons cm^3 / s / keV */
  int     Nrrc;
  double *E_rrc;   /* in keV */
  double *rrc;     /* in photons cm^3 / s / keV */
  int     Nreal;
  double *E_real;  /* in keV */
  double *real;    /* in photons cm^3 / s / keV (sum of brems, twoph, rrc) */
  double *bin_real;/* in ergs cm^3 / s / bin (sum of brems, twoph, rrc) */
  int     Npseudo;
  double *E_pseudo;/* in keV */
  double *pseudo;  /* in photons cm^3 / s / keV */
  struct COMPRESS_CONT *next;
};

struct BIN_SPEC {
  long spec_type;
  long Nbin;
  double bin_min;      /* In keV */
  double bin_max;      /* In keV */
  double bin_logmin;      /* In log keV */
  double bin_logmax;      /* In log keV */
  double delta_bin;      /* In keV */
  double *bin_edge;      /* In keV */
  double total_emission[MAXIMUM_Z+1];  /* In ergs cm^3 / s */
  double tot_line_emis[MAXIMUM_Z+1];   /* In ergs cm^3 / s */
  double tot_brems_emis[MAXIMUM_Z+1];  /* In ergs cm^3 / s */
  double tot_rrc_emis[MAXIMUM_Z+1];    /* In ergs cm^3 / s */
  double tot_twoph_emis[MAXIMUM_Z+1];  /* In ergs cm^3 / s */
  double *bin_emission[MAXIMUM_Z+1];  /* In photons cm^3 / s */
  double *pseudo_emiss[MAXIMUM_Z+1];  /* In photons cm^3 / s */
  double *brems_emiss[MAXIMUM_Z+1];   /* in photons cm^3 / s */
  double *rrc_emiss[MAXIMUM_Z+1];     /* in photons cm^3 / s */
  double *twoph_emiss[MAXIMUM_Z+1];   /* in photons cm^3 / s */
  /* double *real_emiss[MAXIMUM_Z+1]; */   /* In photons cm^3 / s */ 
  /* real_emiss is the sum of the emission per ion  */
  double *bin_error[MAXIMUM_Z+1];
  struct COMPRESS_CONT *cs;  /* Linked list of spectral elements. */
};

/* ******************************************* */


struct ION_BAL_TABLE {
  int nei;      /* If true, this is an NEI ionization balance */
  int Nrows;    /* Total number of rows */
  int Ntemp;    /* Number of temperatures */
  int Ndens;    /* Number of densities */
  float *T;     /* temperature (K) array */
  float *dens;  /* density (cm^-3) array */
  int NZ;       /* actual number of elements in table */
  int Z[MAXIMUM_Z+1];    /* protons (Z) in i-th element */
  int indx[MAXIMUM_Z+1]; /* Index into X array */
  float **X;    /* ionization balance (unitless, normed to unity) */
                /* where X[T,n] = X[T*ndens + n] */
};

/* ****************************************************** */
/* PARAMETERS contains values that are set at the outset  */
/* of the run, and do not change afterward.               */
/* ****************************************************** */

struct PARAMETERS {       /* structure to collect all the plasma parameters */
  char fn_linespec[MAXSTRLEN];
  char fn_cocospec[MAXSTRLEN];
  char fn_compspec[MAXSTRLEN];
  char fn_bincontspec[MAXSTRLEN];
  char fn_xspectable[MAXSTRLEN];
  char modelname[MAXSTRLEN];
  int do_line;
  int do_coco;
  int do_comp;
  int do_bincont;
  int do_xspectable;
  fitsfile *fptr_line;
  fitsfile *fptr_coco;
  fitsfile *fptr_comp;
  fitsfile *fptr_bincont;
  fitsfile *fptr_xspectable;

  char filemap_file[MAXSTRLEN];
  int require_checksum;
  char ionbal_file[MAXSTRLEN];

  char abund_source[MAXSTRLEN];
  int Brems_Type;         /* Bremsstralung gaunt factor calculation method. */

  int do_rrc;    /* If TRUE, calculate radiative recomb. continuum */
  int do_sat;    /* If TRUE, calculate satellite lines */
  int do_lines;  /* If TRUE, calculate line emission */
  int do_twoph;  /* If TRUE, calculate two photon emission */
  int do_brems;  /* If TRUE, calculate bremsstrahlung radiation */

  char inc_atoms[MAXSTRLEN];  /* Input atom string */
  char inc_ions[MAXSTRLEN];   /* Input ion string */
  int N_Elements;/* Number of elements */
  int *list_Z;   /* Array of elements */
  int *list_NrmJ;/* Number of ions used for each element*/
  int *list_rmJ[MAXIMUM_Z+1]; /* Array of ions */

  double min_epsilon;  /* in photons cm^3 / s, min value for photon emission */

  int Ndens;
  int Dens_linear;
  double Dens_Step;
  double Dens_Start;
  double *density;
  int Ntemp;
  int T_linear;
  double T_Step;
  double T_Start;
  double *temperature;
  double Lambda_Start;
  double Lambda_End;
  long BinType;  
  long Nbin;
  double BinMin;
  double BinMax;

  int nei;  /* If true, ion balance is from a table of an nei calculation */
  int nei_output; /* If true, all results are per ion, not per element */
};

/* Checksum storage */

struct CHKSUM {
  char keyname[9];
  char keyval[17];
  int Z;
  int rmJ;
  int type;
  struct CHKSUM *next;
};

/* ******************************************* */

/* FITSDEBUG */
#define DEFAULT_TABLE_TYPE BINARY_TBL
#define IONBAL_READ 1
#define IONBAL_CALC 2
#define DEBUG_DO_ASCII     1
#define DEBUG_DO_FITS      2

struct DEBUG_GENERIC {
  FILE *ascii;
  fd_file fits;
};

struct DEBUG_FILES {
  int ascii;                   /* write ascii files */
  int fits;                    /* write fits files */
  int table_type;              /* ascii or binary fits tables */
  int ionbal_read_or_calc;     /* is ionization balance read or calculated */
  int do_emiss;                /* Emission tables */
  struct DEBUG_GENERIC emiss;
  int do_ionbal;               /* Ionization balance tables */
  struct DEBUG_GENERIC ionbal;
  int do_table;                /* Excitation/de-excitation rate data */
  struct DEBUG_GENERIC table;
  int do_qioniz;               /* Ionization rates */
  struct DEBUG_GENERIC qioniz;
  int do_qrec;                 /* Recombination rates */
  struct DEBUG_GENERIC qrec;
  int do_qtrans;               /* Transition rates */
  struct DEBUG_GENERIC qtrans;
  int do_rates;                /* Other rates */
  struct DEBUG_GENERIC rates;
  int do_satel;                /* Satellite lines */
  struct DEBUG_GENERIC satel;  
  int do_twoph;                /* Two photon emission */
  struct DEBUG_GENERIC twoph;  
  int do_levpop;               /* Level populations */
  struct DEBUG_GENERIC levpop;  
  int do_compress;             /* Compression calculations */
  struct DEBUG_GENERIC compress;  
};
