#include <stdarg.h>
#include "fitsio.h"
#include "utilities.h"

#ifndef _FITSDEBUG_H
#define _FITSDEBUG_H

#ifndef MAXSTRLEN
#define MAXSTRLEN 128
#endif

#define FD_VERBOSE 2      /* Verbosity level of fitsdebug library */
                          /* FD_VERBOSE = 0 : completely quiet    */ 
                          /* FD_VERBOSE = 1 : fairly quiet        */ 
                          /* FD_VERBOSE = 2 : lots of diagnostics */ 

#define FD_MAXCOL 100
#define FD_VALID_EXT_TYPES {"TABLE", "BINTABLE"}
#define FD_DEFAULT_COMMENT "This file has been created using the fitsdebug library."

/* Structure */
#define FD_FILE_OPEN      1
#define FD_FILE_CLOSED    0
#define FD_FILE_UNKNOWN  -1
#define FD_SUPERSEDE_ON   1
#define FD_SUPERSEDE_OFF  0
#define FD_FLUSH_ON       1
#define FD_FLUSH_OFF      0
#define FD_ROW_START     -101
#define FD_ROW_END       -102

typedef struct TYPECODE {
  char *raw;
  char *type_code;
  char type;
  int  int_type;
  int size;
  int precision;
  int vla;
  int vla_length;
} typecodeS;



typedef struct FD_FILE_STRUCT {
  fitsfile *fptr;                  /* pointer to a fitsfile      */
  char filename[MAXSTRLEN];        /* name of the fitsfile       */
  char tablename[MAXSTRLEN];       /* name of the table          */
  int tabletype;                   /* extension type             */
  int hdu;                         /* HDU number if the table    */
  int ncol;                        /* number of columns          */
  long nrow;                       /* number of rows             */
  typecodeS *typecode;
  int *tbcol;
  char **tform;
  char **ttype;
  char **tunit;
  int flush;                       /* flush file or open/close   */
  int supersede;                   /* supersede on creation?     */
  int file_status;                 /* file status                */
                                   /*   open   : FD_FILE_OPEN    */
                                   /*   closed : FD_FILE_CLOSED  */
                                   /*   ???    : FD_FILE_UNKNOWN */
} fd_file;


typedef union {
    unsigned char *byt_val;
    char         **str_val;
    /* logical    *log_val; */
    short         *sht_val;
    int           *int_val;
    long          *lng_val;
    float         *flt_val;
    double        *dbl_val;
    /* complex    *cmp_val;
       doublecomplex *dblcmp_val; */
  } darray;

/* Error/Status codes */
#define FD_STAT_ALLOK          0 /* all actions were handle without problems */
#define FD_STAT_SUPERSEDE   1001 /* supersede in fd_create_file does not have
				    the value FD_SUPERSEDE_ON or 
				              FD_SUPERSEDE_OFF */
#define FD_STAT_FLUSH       1002 /* flush in fd_create_file does not have
				    the value FD_FLUSH_ON or 
				              FD_FLUSH_OFF */
#define FD_STAT_FILE_NOT_OPEN 1003 /* fits file is not open */
#define FD_STAT_TABLE_TYPE  1004 /* Table type specified is not available */
                                 /* in cfitsio library                    */ 
#define FD_STAT_NO_INT      1005 /* int type declaration is not allowed   */
#define FD_STAT_NO_TYPE     1006 /* no sensible type declaration provided */
#define FD_STAT_NCOL        1007 /* incorrect number of arguments to      */
                                 /* fd_create_file                        */


/* Function definitions */
int fd_create_file(fd_file *file, char *filename, 
		   int tabletype, char *tablename, 
		   int *status, int flush, int supersede, 
		   int ncol, ...);
void fd_print_fd_file(fd_file *file);
int  fd_close_file(fd_file *file, int *status);
void fd_err_message(int status);
int  fd_file_exist(char *filename);
int  fd_append_extension(char *filename);
char *fd_get_typecode_binary(typecodeS *result, char *type, int *status);
char *fd_get_typecode_ascii(typecodeS *result, char *type, int *status);
int fd_set_typecode_binary(typecodeS *typecode, char* tform, int *status);
int fd_set_typecode_ascii(typecodeS *typecode, char* tform, int *status);
void fd_print_typecode(typecodeS typecode);
int fd_get_type_from_typecode(typecodeS typecode);
int fd_get_size_from_typecode(typecodeS typecode);
long fd_get_nelem_from_typecode(typecodeS typecode);
int fd_exit_file(fd_file *file, int *status);
int fd_setup_table(fd_file *file, int ncol, int *status);
int fd_flush_close(fd_file *file);
int fd_flush_open(fd_file *file);
int fd_write_row(fd_file *file, ...);
int fd_vwrite_row(fd_file *file, va_list *args);
int fd_write_rows(fd_file *file, long nrows, ...);
void fd_darray_malloc(darray *data, typecodeS typecode, long nelem);
void fd_darray_free(darray *data, typecodeS typecode, long nelem);
void fd_darray_value_1row(darray *data, typecodeS typecode, long nelem, va_list *ap);
void fd_darray_value_nrow(darray *data, typecodeS typecode, long nelem, va_list *ap);
void fd_darray_print(darray *data, typecodeS typecode, long nelem);
int fd_darray_write_col(fd_file *file, darray *data, typecodeS typecode, 
			int col, long firstrow, long firstelem,
			long nelem, int *status);
int fd_print_rows(FILE *output, fd_file *file, long row_start, long row_end);
int fd_print_col(fd_file *file, int col, char **final, 
		 long row_start, long nrows);
int fd_print_col_binary(fd_file *file, int col, char **final, 
			long row_start, long nrows);
int fd_print_col_ascii(fd_file *file, int col, char **final, 
		       long row_start, long nrows);
int fd_get_print_fsize(fd_file *file);
int fd_open_file(fd_file *file, char *filename, 
		 char *tablename, int *status, int flush);
int fd_comment(fd_file *file, char *comment, int *status);
int fd_comment_no_flush(fd_file *file, char *comment, int *status);
int fd_set_ckeyword(fd_file *file, char *keyword, 
		    char *value, char *comment, int *status);
int fd_set_ckeyword_no_flush(fd_file *file, char *keyword, 
			     char *value, char *comment, int *status);
int fd_set_ikeyword(fd_file *file, char *keyword, 
		    int value, char *comment, int *status);
int fd_set_ikeyword_no_flush(fd_file *file, char *keyword, 
			     int value, char *comment, int *status);
int fd_set_fkeyword(fd_file *file, char *keyword, 
		    float value, char *comment, int *status);
int fd_set_fkeyword_no_flush(fd_file *file, char *keyword, 
			     float value, char *comment, int *status);

int fd_set_fptr(fd_file *file, fitsfile *fptr);
fitsfile *fd_get_fptr(fd_file *file);
int fd_set_filename(fd_file *file, char filename[MAXSTRLEN]);
char *fd_get_filename(fd_file *file);
int fd_set_tablename(fd_file *file, char tablename[MAXSTRLEN]);
char *fd_get_tablename(fd_file *file);
int fd_set_tabletype(fd_file *file, int tabletype);
int fd_get_tabletype(fd_file *file);
int fd_set_hdu(fd_file *file, int hdu);
int fd_get_hdu(fd_file *file);
int fd_set_ncol(fd_file *file, int ncol);
int fd_set_ncol_ap(fd_file *file, va_list ap);
int fd_get_ncol(fd_file *file);
int fd_set_nrow(fd_file *file, long nrow);
long fd_get_nrow(fd_file *file);
int fd_set_flush(fd_file *file, int flush);
int fd_get_flush(fd_file *file);
int fd_set_supersede(fd_file *file, int supersede);
int fd_get_supersede(fd_file *file);
int fd_set_file_status(fd_file *file, int file_status);
int fd_get_file_status(fd_file *file);
int fd_set_tformi(fd_file *file, int i, char *tform);
char *fd_get_tformi(fd_file *file, int i);
int fd_set_ttypei(fd_file *file, int i, char *tform);
char *fd_get_ttypei(fd_file *file, int i);
int fd_set_tuniti(fd_file *file, int i, char *tform);
char *fd_get_tuniti(fd_file *file, int i);

void fd_print_tform(fd_file *file);
void fd_print_ttype(fd_file *file);
void fd_print_tunit(fd_file *file);

#endif        /* End of #define _FITSDEBUG_H */
