/** \file Application.h
    \brief Declaration of Application class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <iosfwd>
#include <list>
#include <string>

#ifndef profit_Application_h
#define profit_Application_h

namespace profit {

  /** \class Application
      \brief Class encapsulating the main profit application.
  */
  class Application {
    public:
      /** \brief Parse the command line, determining whether to run in debug mode and/or test mode, etc.
          \param argc The number of arguments.
          \param argv The arguments.
      */
      static void parseCommandLine(int argc, char ** argv);

      /// \brief Create an application of the type previously determined by the command line arguments.
      static Application * create();

      /// \brief Force test mode to the given value, overriding command line.
      static void setTestMode(bool test_mode);

      /// \brief Destroy this application.
      virtual ~Application();

      /// \brief Perform the action required by this application.
      virtual void run();

      /// \brief Returns flag indicating whether the application failed to execute.
      virtual bool failed() const;

      /// \brief Return a string describing the version number of this release.
      virtual const std::string & version() const;

    protected:
      /** \brief Helper which creates the actual application without catching exceptions.
      */
      static Application * createApp();

      /** \brief Return a stream for reporting errors.
          \param method The method which is writing the message. Used for formatting.
      */
      static std::ostream & error(const std::string & method);

      /** \brief Return an info stream.
          \param method The method which is writing the message. Used for formatting.
      */
      static std::ostream & info(const std::string & method);

      /** \brief Return an warning stream.
          \param method The method which is writing the message. Used for formatting.
      */
      static std::ostream & warn(const std::string & method);

      /// \brief Create the application. Only allowed in subclasses.
      Application();

      /// \brief Perform the detailed action apppropriate for this application without catching exceptions.
      virtual void runApp() = 0;

      static std::string s_application_name;
      static std::list<std::string> s_arg;
      static bool s_debug_mode;
      static bool s_gui_mode;
      static bool s_hit_mode;
      static bool s_test_mode;
      static bool s_failed;
      static bool s_version_mode;
      std::string m_version;

    private:
      /// \brief Copy the application. This is not allowed; it is a singleton.
      Application(const Application &);
  };

}

#endif
