/** \file AtomDb.h
    \brief Declaration of AtomDb class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_AtomDb_h
#define profit_AtomDb_h

#include "profit/AtomicLine.h"

#include <string>
#include <vector>

namespace profit {

  class AtomDbImp;
  class IEnergy;

  /** \class AtomDb
      \brief Class encapsulating access to the atomic database.
  */
  class AtomDb {
    public:
      typedef std::vector<AtomicLine> AtomicLineCont;

      /** \brief Create an atomic database accessor object.
          \param file_name The name of the atomic database file.
      */
      AtomDb(const std::string & file_name);

      ~AtomDb();

      /** \brief Discard database entries (in memory) whose energy falls outside an energy range.
          \param min_energy Energy object with minimum value of range.
          \param max_energy Energy object with maximum value of range.
      */
      void filter(const IEnergy & min_energy, const IEnergy & max_energy);

      /** \brief Discard database entries (in memory) whose emissivity is less than the given threshold.
          \param min_emissivity Minimum emissivity accepted.
      */
      void filter(double min_emissivity);

      /// \brief Get container of atomic lines which match the current criteria for line selection.
      const AtomicLineCont & getMatches() const;

    private:
      AtomDbImp * m_imp;
  };

}

#endif
