/** \file AtomicLevel.h
    \brief Declaration of AtomicLevel class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_AtomicLevel_h
#define profit_AtomicLevel_h

#include <string>

/// \struct LEVEL_DATA
/// \brief Forward declaration for data structure defined in APEC used internally by AtomicLevel and associated classes.
struct LEVEL_DATA;

namespace profit {

  /** \class AtomicLevel
      \brief Encapsulation of a single atomic quantum state.
  */
  class AtomicLevel {
    public:
      /// \brief Create an empty atomic level, that is, one containing no information.
      AtomicLevel();

      /** \brief Create an atomic level asociated with the given internal APEC object.
          \param level_data The APEC data object containing all information available for this level.
      */
      AtomicLevel(const LEVEL_DATA & level_data);

      /// \brief Return the energy of this state, in keV.
      double energy() const;

      /// \brief Return the principle quantum number, N.
      int quantumN() const;

      /// \brief Return the orbital angular momentum quantum number, l.
      int quantumL() const;

      /// \brief Return the spin angular momentum quantum number, s.
      double quantumS() const;

      /// \brief Return the degeneracy of this level.
      int degeneracy() const;

      /// \brief Return the level number of this level.
      int level() const;

      /// \brief Construct a string describing the configuration.
      std::string config() const;

      /// \brief Construct a string describing the configuration.
      std::string description() const;

    private:
      const LEVEL_DATA * m_level_data;
  };

}

#endif
