/** \file AtomicLine.h
    \brief Declaration of AtomicLine class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_AtomicLine_h
#define profit_AtomicLine_h

#include <string>

namespace profit {

  /** \class AtomicLine
      \brief Encapsulation of all information associated with one atomic line level.
  */
  class AtomicLine {
    public:
      /// \brief Create AtomicLine object with empty atomic line properties.
      AtomicLine();

      /** \brief Create AtomicLine object with the given atomic line properties.
          \param database The name of the database from which this line was obtained.
          \param lambda The wavelength in Angstroms.
          \param element The atomic element number (Z).
          \param ion The ionization (number of electrons in the configuration).
          \param upper_level The upper level of the transition.
          \param lower_level The lower level of the transition.
          \param peak_emissivity The peak emissivity, in photons-cm^3 / s.
      */
      AtomicLine(const std::string & database, double lambda, int element, int ion, int upper_level, int lower_level,
        double peak_emissivity);

      /// \brief Return the wavelength, in Angstroms.
      double getLambda() const;

      /// \brief Return the element number.
      int getElement() const;

      /// \brief Return the ion number.
      int getIon() const;

      /// \brief Return the upper level of the transition.
      int getUpperLevel() const;

      /// \brief Return the lower level of the transition.
      int getLowerLevel() const;

      /// \brief Return the peak emissivity, in photons-cm^3 / s.
      double getPeakEmissivity() const;

      /** \brief Create a string describing this line, in the form.
          \param reference_line Another atomic line object, used to format peak emissivity as a fractional
          value relative to the reference line's peak emissivity.
          \param wavelength_space If true, description will report wavelength, otherwise energy.
      */
      std::string getLabel(const AtomicLine & reference_line, bool wavelength_space) const;

      /** \brief Create a string containing the name of this line, using conventional roman numbering
          for the element and ionization. */
      std::string getRomanName() const;

      const std::string & database() const;

      bool operator ==(const AtomicLine & line) const;

    private:
      std::string m_database;
      double m_lambda;
      double m_peak_emissivity;
      int m_element;
      int m_ion;
      int m_upper_level;
      int m_lower_level;
  };

}

#endif
