/** \file AtomicTransition.h
    \brief Declaration of AtomicTransition class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_AtomicTransition_h
#define profit_AtomicTransition_h

#include <list>

#include "profit/AtomicLevel.h"

// Forward declarations for data structures defined in APEC used internally by AtomicTransition and associated classes.
struct COLLTRNS_DATA;
struct LINERAD_DATA;
struct SATELLITE_DATA;

namespace profit {

  class OStream;

  /** \class AtomicTransition
      \brief Class encapsulating a transition between two atomic states, with details of all processes which
      are known to contribute to the transition rate.
  */
  class AtomicTransition {
    public:
      // Provisional typedefs which wrap APEC types. May end up becoming classes.
      typedef COLLTRNS_DATA * CollProcess;
      typedef LINERAD_DATA * RadProcess;
      typedef SATELLITE_DATA * SatProcess;

      // Typedefs for convenience.
      /// \typedef CollCont Container of CollProcess objects.
      typedef std::list<CollProcess> CollCont;
      /// \typedef CollCont Container of RadProcess objects.
      typedef std::list<RadProcess> RadCont;
      /// \typedef CollCont Container of SatProcess objects.
      typedef std::list<SatProcess> SatCont;

      /// \brief Create an atomic transition object with no processes and undefined upper and lower level states.
      AtomicTransition();

      /** \brief Create an atomic transition object with no processes.
          \param Z The atomic number.
          \param rmJ
          \param lower_level The lower energy state.
          \param upper_level The lower energy state.
      */
      AtomicTransition(int Z, int rmJ, AtomicLevel lower_level, AtomicLevel upper_level);

      /// \brief Destruct an atomic transition object.
      virtual ~AtomicTransition();

      /** \brief Add a CollProcess to the set of processes associated with this transition.
          \param process The CollProcess object to add.
      */
      void addProcess(const CollProcess & process);

      /** \brief Add a RadProcess to the set of processes associated with this transition.
          \param process The RadProcess object to add.
      */
      void addProcess(const RadProcess & process);

      /** \brief Add a SatProcess to the set of processes associated with this transition.
          \param process The SatProcess object to add.
      */
      void addProcess(const SatProcess & process);

      /// \brief Accessor returning the lower atomic level object for this transtion.
      const AtomicLevel & getLowerLevel() const;

      /// \brief Accessor returning the upper atomic level object for this transtion.
      const AtomicLevel & getUpperLevel() const;

      /// \brief Accessor returning the container of CollProcess objects.
      const CollCont & getCollision() const;

      /// \brief Accessor returning the container of RadProcess objects.
      const RadCont & getRadiative() const;

      /// \brief Accessor returning the container of SatProcess objects.
      const SatCont & getSatellite() const;

      /** \brief Write summary of this atomic transition to an output stream.
          \param os The stream to which to write the state.
          \param wavelength_space Flag indicating whether to report wavelengths or energies of transitions.
      */
      virtual OStream & write(OStream & os, bool wavelength_space = true) const;

    private:
      AtomicLevel m_lower_level;
      AtomicLevel m_upper_level;
      CollCont m_coll;
      RadCont m_rad;
      SatCont m_sat;
      int m_Z;
      int m_rmJ;
  };

  /** \brief Standard shift operator for writing AtomicTransition to an output stream.
      \param os The stream to which to write.
      \param transition The transition to write.
  */
  OStream & operator <<(OStream & os, const AtomicTransition & transition);

}

#endif
