/** \file DataEntryWidget.h
    \brief Declaration of DataEntryWidget class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_DataEntryWidget_h
#define profit_DataEntryWidget_h

#include <list>
#include <sstream>
#include <stdexcept>
#include <string>

#include "RQ_OBJECT.h"
#include "Rtypes.h"
#include "TGLayout.h"
#include "TQObject.h"

// RootCint is finicky if the header files themselves are not included.
#include "TGButton.h"
#include "TGFrame.h"
#include "TGLabel.h"
#include "TGTextEntry.h"

namespace profit {

  /** \class DataEntryWidget
      \brief Class encapsulating a widget for data entry.
  */
  class DataEntryWidget {
    public:
      RQ_OBJECT("profit::DataEntryWidget")
      ClassDef(DataEntryWidget, 1)

      DataEntryWidget(const std::string & style, const std::string & label, const std::string & content = "");

      virtual ~DataEntryWidget();

      virtual void buildGui(TGCompositeFrame * parent, unsigned int width, unsigned int height);

      virtual void handleClicked();

      virtual void handleTextEntry(const char * content);

      template <typename T>
      void get(T & value) const;

      template <typename T>
      void set(const T & value);

      void clearDisplay();

      // Temporary hack to allow buttons to be in mutually exclusive configurations.
      void tie(DataEntryWidget * widget);

      void control(DataEntryWidget * widget);

      void activate(bool is_on);

    private:
      // Do not use this constructor; it is only needed by RootCint.
      DataEntryWidget();

      std::list<DataEntryWidget *> m_controlled;
      std::string m_style;
      std::string m_label_text;
      std::string m_content;
      TGLayoutHints m_centerY_expandX;
      TGLayoutHints m_centerY_pad_all;
      TGCompositeFrame * m_parent;
      TGHorizontalFrame * m_frame;
      TGLabel * m_label;
      TGFrame * m_widget;
      TGButton * m_button;
      TGTextEntry * m_text_entry;
      DataEntryWidget * m_tied_widget;
      bool m_activated;
  };

  template <typename T>
  inline void DataEntryWidget::get(T & value) const {
    std::stringstream ss;
    ss << m_content;
    T tmp_value;
    ss >> tmp_value;
    if (!ss.eof())
      throw std::runtime_error("DataEntryWidget::get(T & value): could not convert string " + m_content + " to type T");
    value = tmp_value;
  }

  template <typename T>
  inline void DataEntryWidget::set(const T & value) {
    std::ostringstream os;
    os << value;
    m_content = os.str();
    if (0 != m_button) {
      m_button->SetOn((m_content != "off" && m_content != "0") ? kTRUE : kFALSE);
      for (std::list<DataEntryWidget *>::iterator itor = m_controlled.begin(); itor != m_controlled.end(); ++itor) {
        (*itor)->activate((m_content != "off" && m_content != "0") ? true: false);
      }
    } else if (0 != m_text_entry) {
      m_text_entry->SetText(m_content.c_str());
    }
  }

}

#endif
