/** \file Energy.h
    \brief Declarations for energy-related abstractions.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_Energy_h
#define profit_Energy_h

namespace profit {

  /** \class IEnergy
      \brief Base class encapsulating the idea of an energy, which may be manifested in different representations.
      in subclasses.
  */
  class IEnergy {
    public:
      //  Product of Planck's constant and speed of light, in keV-Angstroms, used to convert energy to wavelength and vice versa.
      static const double s_h_c;

      // Small but non-0 value to protect from division by 0.
      static const double s_epsilon;
      static const double s_min_energy;
      static const double s_max_energy;
      static const double s_min_wavelength;
      static const double s_max_wavelength;

      /// \brief Destroy an IEnergy object.
      virtual ~IEnergy() {}

      /** \brief Assignment from another IEnergy.
          \param ienergy The source IEnergy object.
      */
      virtual IEnergy & operator =(const IEnergy & ienergy) = 0;

      /** Return energy in keV. */
      virtual double energy() const = 0;

      /** Return wavelength in Angstroms. */
      virtual double wavelength() const = 0;

      /** \brief Test for equality.
          \param energy The energy being compared to this one.
      */
      virtual bool operator ==(const IEnergy & ienergy) const { return energy() == ienergy.energy(); }

      /** \brief Test for inequality.
          \param energy The energy being compared to this one.
      */
      virtual bool operator !=(const IEnergy & ienergy) const { return energy() != ienergy.energy(); }
  };

  /** \class Energy
      \brief An IEnergy subclass in which energy is represented directly as a single energy value.
  */
  class Energy : public IEnergy {
    public:
      /** \brief Create an Energy object with the given energy.
          \param energy The energy, in keV.
      */
      Energy(double energy): m_energy(energy < s_min_energy ? s_min_energy :
        (energy > s_max_energy ? s_max_energy : energy)) {}

      /** \brief Construct an Energy from an IEnergy object, converting to the correct internal representation.
          \param ienergy The source IEnergy object.
      */
      Energy(const IEnergy & ienergy): m_energy(ienergy.energy()) {}

      /** \brief Assign to this Energy from an IEnergy object, converting to the correct internal representation.
          \param ienergy The source IEnergy object.
      */
      virtual IEnergy & operator =(const IEnergy & ienergy) { m_energy = ienergy.energy(); return *this; }

      /** \brief Assign to this Energy from another Energy object.
          \param energy The source Energy object.
      */
      Energy & operator =(const Energy & energy) { m_energy = energy.m_energy; return *this; }

      /** Return energy in keV. */
      virtual double energy() const { return m_energy; }

      /** Return wavelength in Angstroms. */
      virtual double wavelength() const { return s_h_c / m_energy; }

    private:
      double m_energy;
  };

  /** \class Wavelength
      \brief An IEnergy subclass in which energy is represented indirectly as a single wavelength value.
  */
  class Wavelength : public IEnergy {
    public:
      /** \brief Create a Wavelength object with the given wavelength.
          \param wavelength The wavelength, in Angstroms.
      */
      Wavelength(double wavelength): m_wavelength(wavelength < s_min_wavelength ? s_min_wavelength :
        (wavelength > s_max_wavelength ? s_max_wavelength : wavelength)) {}

      /** \brief Construct a Wavelength from an IEnergy object, converting to the correct internal representation.
          \param ienergy The source IEnergy object.
      */
      Wavelength(const IEnergy & energy): m_wavelength(energy.wavelength()) {}

      /** \brief Assign to this Wavelength from an IEnergy object, converting to the correct internal representation.
          \param ienergy The source IEnergy object.
      */
      virtual IEnergy & operator =(const IEnergy & ienergy) { m_wavelength = ienergy.wavelength(); return *this; }

      /** \brief Assign to this Wavelength from another Wavelength object.
          \param wavelength The source Wavelength object.
      */
      Wavelength & operator =(const Wavelength & wavelength) { m_wavelength = wavelength.m_wavelength; return *this; }

      /** Return energy in keV. */
      virtual double energy() const { return s_h_c / m_wavelength; }

      /** Return wavelength in Angstroms. */
      virtual double wavelength() const { return m_wavelength; }

    private:
      double m_wavelength;
  };

}

#endif
