/** \file Gui.h
    \brief Declaration of Gui class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_Gui_h
#define profit_Gui_h

#include <list>
#include <string>
#include <vector>

#include "RQ_OBJECT.h"
#include "Rtypes.h"
#include "TGLayout.h"
#include "TQObject.h"
#include "TRootEmbeddedCanvas.h"

class TGFrame;
class TGListBox;
class TGMainFrame;
class TGPopupMenu;
class TGTextButton;
class TGTextView;
class TGVerticalFrame;

#include "profit/AtomDb.h"
#include "RootFrame.h"

namespace profit {

  class ActiveCanvas;
  class InputTable;
  class SpecFile;
  class Spectrum;
  class SpectrumView;
  class StoredCanvas;
  class GuiOStream;
  class UnitTestGui;

  /** \class Gui
      \brief Class encapsulating the top-level profit GUI main frame.
  */
  class Gui {
    public:
      RQ_OBJECT("profit::Gui")
      ClassDef(Gui, 0)

      typedef std::vector<SpecFile *> SpecFileCont;

      friend class UnitTestGui;

      /// \brief Perform global Root initializations.
      static void initRoot();

      /// \brief Construct the GUI singleton.
      static Gui * instance();

      /// \brief Destroy a GUI object.
      virtual ~Gui();

      /// \brief Run the GUI event loop.
      virtual void run();

      /// \brief Close the main GUI window.
      virtual void handleCloseWindow();

      /// \brief Handle line id event.
      virtual void handleLineId();

      /// \brief Handle line selection event.
      virtual void handleLineSelect(int = 0, int = 0);

      /// \brief Handle wavelength view event.
      virtual void handleWavelengthView();

      /// \brief Handle energy view event.
      virtual void handleEnergyView();

      /// \brief Handle velocity view event.
      virtual void handleVelocityView();

      /// \brief Handle fit line view event.
      virtual void handleFitLine();

      /// \brief Handle store spectrum view event.
      virtual void handleStoreView();

      /// \brief Handle atomic info view event.
      virtual void handleAtomicInfo();

      /// \brief Handle atomic info view event.
      virtual void handleLabelLines();

      /// \brief Handle atomic info view event.
      virtual void handleLabelPeaks();

      /// \brief Handle clear line labels event.
      virtual void handleClearLabels();

      /// \brief Handle expand X event.
      virtual void handleExpandX();

      /** \brief Handle menu selection event.
          \param id Identity of the menu item selected.
      */
      virtual void handleMenu(int id);

      virtual void handleSpectrumMenu(int id);

      /** \brief Display clone of the given view in the main view area. Display energy view of the whole
          spectrum in the thumbnail area.
          \param view The view to select/display.
      */
      virtual void handleSelectView(const SpectrumView & view);

      virtual void handleStripChart();

      virtual void handleChartChecked();

      virtual void handleSmooth();

      virtual void handlePreferences();

      virtual void handleAboutProfit();

      /// \brief Display a dialog box to open a spectrum file.
      virtual void openSpecFile(std::string file_name = "", std::string dir_name = "", bool force_dialog = false);
      
      /// \brief Print the currently selected canvas to a file.
      virtual void printMainCanvas();
      
      /// \brief Print the given canvas to a file.
      virtual void printCanvas(const TRootEmbeddedCanvas * canvas) const;
      
      /// \brief Clear line identity list box, and erase matching atomic line information.
      void clearLineId();

      /// \brief Return a reference to the Gui's main message window.
      GuiOStream & getMainOutput();

      /// \brief Set the version string reported by the GUI's "about" dialog box.
      void setVersion(const std::string & version);

    private:
      Gui();

      /// \brief Display a dialog box to open a spectrum.
      virtual void openSpectrum(SpecFile & spec_file, int spec_id);
      
      virtual void displaySpecFileSummary(const SpecFile & spec_file);

      /// \brief Destroy all Root frame objects.
      void deleteAllFrames();

      /// \brief Destroy all SpectrumView objects.
      void deleteAllViews();

      /** \brief Utility to track the given frame pointer in the container of frames for garbage collection.
          Throws exception if pointer is 0, indicating memory allocation failed.
          \param context A string indicating which object is being stored, used to format exception.
          \param frame The frame being stored.
      */
      void trackFrame(const std::string & context, TGFrame * frame, TGCompositeFrame * parent = 0, ULong_t hints = kLHintsNormal,
        Int_t pad_left = 0, Int_t pad_right = 0, Int_t pad_top = 0, Int_t pad_bottom = 0);

      /** \brief Utility to delete the given frame pointer, and remove it from the container of frames.
          \param frame The frame being deleted.
      */
      void deleteFrame(TGFrame * frame);

      /** \brief Utility to track the given view pointer in the container of views for garbage collection.
          Throws exception if pointer is 0, indicating memory allocation failed.
          \param context A string indicating which object is being stored, used to format exception.
          \param view The view being stored.
      */
      void trackView(const std::string & context, SpectrumView * view);

      /** \brief Utility to delete the given view pointer, and remove it from the container of views.
          \param view The view being deleted.
      */
      void deleteView(SpectrumView * view);

      /// \brief Construct menu in top level main frame.
      virtual void constructMenu();

      /// \brief Construct button bars which divide top from bottom in main frame.
      virtual void constructButtonBars();

      /** \brief Display a file dialog box and return the value selected by the user.
          \param file_types Array of pairs of strings giving description and file glob for determining
          what types of files to display.
          \param init_file_name The initial file name displayed in the dialog box.
          \param init_dir_name The name of the initial directory displayed in the dialog box.
      */
      virtual std::string fileDialog(const char ** file_types, const std::string & init_file_name = "",
        const std::string & init_dir_name = "") const;

      std::list<RootFrame> m_frame_cont;
      std::list<SpectrumView *> m_view_cont;
      std::list<SpectrumView *> m_stored_view;
      std::vector<int> m_spectrum_index;
      SpecFileCont m_spec_file_cont;
      AtomDb::AtomicLineCont m_matches;
      std::string m_version;
      std::string m_atom_db_file_name;
      std::string m_xstar_db_file_name;
      std::string m_spec_file_name;
      std::string m_spec_dir_name;
      TGLayoutHints m_lh_pad;
      TGLayoutHints m_lh_expand;
      TGLayoutHints m_lh_expand_x;
      TGLayoutHints m_lh_expand_y;
      TGLayoutHints m_lh_expand_pad;
      TGLayoutHints m_lh_expand_x_pad;
      double m_emissivity_threshold;
      double m_src_redshift;
      double m_min_velocity;
      double m_max_velocity;
      unsigned int m_number_plots;
      unsigned int m_spec_menu_size;
      TGMainFrame * m_main_frame;
      TGTextView * m_spectrum_summary;
      TGTextButton * m_wavelength_button;
      TGTextButton * m_energy_button;
      TGTextButton * m_velocity_button;
      TGTextButton * m_atomic_info;
      StoredCanvas * m_small_canvas;
      ActiveCanvas * m_large_canvas;
      TGTextView * m_text_output;
      TGListBox * m_line_box;
      TGCanvas * m_store_canvas;
      TGVerticalFrame * m_store;
      SpecFile * m_spec_file;
      Spectrum * m_spectrum;
      SpectrumView * m_thumbnail_view;
      SpectrumView * m_main_view;
      GuiOStream * m_main_output;
      TGPopupMenu * m_spectrum_menu;
      bool m_use_aped;
      bool m_use_xstar;
      bool m_wavelength_space;
      bool m_const_velocity_bounds;
  };

  class ActiveCanvas : public TRootEmbeddedCanvas {
    public:
      ActiveCanvas(const char * name = "0", const TGWindow * p = 0, UInt_t w = 10, UInt_t h = 10);

      virtual Bool_t HandleContainerButton(Event_t *);
      virtual Bool_t HandleContainerMotion(Event_t *);

      virtual void setView(SpectrumView * view);

    private:
      profit::SpectrumView * m_view;
      double m_x;
      bool m_button_3_down;
  };

  class StoredCanvas : public TRootEmbeddedCanvas {
    public:
      StoredCanvas(const char * name = "0", const TGWindow * p = 0, UInt_t w = 10, UInt_t h = 10);

      virtual Bool_t HandleContainerButton(Event_t * event);
      virtual Bool_t HandleContainerDoubleClick(Event_t *);

      virtual void setView(SpectrumView * view);

    private:
      profit::SpectrumView * m_view;
  };

}

#endif
