/** \file SpecFile.h
    \brief Declaration of SpecFile class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/

#ifndef profit_SpecFile_h
#define profit_SpecFile_h

#include <string>
#include <vector>

namespace profit {

  class InputTable;
  class Spectrum;

  /** \class SpecFile
      \brief Class representing an entire file which contains one or more spectra.
  */
  class SpecFile {
    public:
      typedef std::vector<Spectrum *> SpecCont;
      typedef SpecCont::size_type size_type;

      /** \brief Create a spectrum file object valid for the given file.
          \param spec_file_name The name of the file, which at a minimum must contain a spectrum extension.
      */
      SpecFile(const std::string & spec_file_name);

      virtual ~SpecFile();

      virtual size_type getNumSpectra() const;

      virtual Spectrum * getSpectrum(int spec_index = 0);

      virtual Spectrum * getSpectrum(const std::string & ebounds_file_name, int spec_index = 0);

      virtual std::string getFileName() const;

      virtual void getSummary(std::vector<std::string> & summary) const;

      virtual const InputTable * getInputTable() const;

    private:
      /** \brief Format a string which contains a number to a specified precision level.
          \param number The input numeric string.
          \param prec the precision to use when formatting the output.
      */
      std::string formatNumericString(const std::string & number, int prec) const;

      SpecCont m_spec_cont;
      std::string  m_spec_file_name;
      InputTable * m_spec_table;
  };

}

#endif

