/** \file Spectrum.h
    \brief Declaration of Spectrum class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_Spectrum_h
#define profit_Spectrum_h

#include <string>
#include <vector>

namespace profit {

  class InputTable;
  class SpecFile;

  /** \class Spectrum
      \brief Class encapsulating a single spectrum.
  */
  class Spectrum {
    public:
      /** \brief Create a Spectrum object, reading the spectrum and energy bounds from the given table.

          If the spectrum table has no e_max field, each row of e_max will be computed from e_min
          of the preceding row. If the spectrum table also has no e_min an exception will be thrown.
          \param spectrum_table The table containing counts and channel number (SPECTRUM extension in a PHA file).
      */
      Spectrum(const SpecFile & spectrum_table, int spec_index = -1);

      /** \brief Create a Spectrum object, reading the spectrum from the given input tables.
          \param spectrum_table The table containing counts and channel number (SPECTRUM extension in a PHA file).
          \param ebounds_table The table containing the apparent energy bin definitions (EBOUNDS extension).
      */
      Spectrum(const SpecFile & spectrum_table, const InputTable & ebounds_table);

      /** \brief Create a Spectrum object, reading the spectrum from the given input tables.
          \param spectrum_table The table containing counts and channel number (SPECTRUM extension in a PHA2 file).
          \param ebounds_table The table containing the apparent energy bin definitions (EBOUNDS extension).
          \param spec_index The index identifying which spectrum in the PHA2 file to open.
      */
      Spectrum(const SpecFile & spectrum_table, const InputTable & ebounds_table, int spec_index);

      /** \brief Apply an energy cut to the data contained in this spectrum. This cut does not affect
          the file from which the data were read.
          \param min_energy Minimum energy to accept, in keV.
          \param e_maximum Maximum energy to accept, in keV.
      */
      void select(double min_energy, double max_energy);

      /// \brief Accessor returning the container of lower bounds of currently selected energies, in keV.
      const std::vector<double> & minEnergy() const;

      /// \brief Accessor returning the container of upper bounds of currently selected energies, in keV.
      const std::vector<double> & maxEnergy() const;

      /// \brief Accessor returning the container of counts.
      const std::vector<double> & counts() const;

      /** \brief Attempt to load the spectrum based solely on the information in the file for which
          this configuration is valid.
      */
      void load();

      /** \brief Attempt to load the spectrum based solely on the information in the file for which
          this configuration is valid.
          \param ebounds_file_name Name of file containing ebounds extension.
      */
      void load(const std::string & ebounds_file_name);

      std::string getFileName() const;

      const SpecFile * getSpecFile() const;

    private:
      /** \brief Populate internal data members, reading the data from the given input tables.
          \param spectrum_table The table containing counts and channel number (SPECTRUM extension in a PHA file).
      */
      void load(const InputTable & spectrum_table);

      /** \brief Populate internal data members, reading the data from the given input tables.
          \param spectrum_table The table containing counts and channel number (SPECTRUM extension in a PHA file).
          \param ebounds_table The table containing the apparent energy bin definitions (EBOUNDS extension).
      */
      void load(const InputTable & spectrum_table, const InputTable & ebounds_table);

      /** \brief Populate energy bounds data members, reading the data from the given ebounds tables.
          \param ebounds_table The table containing the apparent energy bin definitions (EBOUNDS extension).
      */
      void loadEnergyBounds(const InputTable & ebounds_table);

      const SpecFile * m_spec_file;
      std::vector<double> m_e_min;
      std::vector<double> m_e_max;
      std::vector<double> m_counts;
      int m_spec_index;
  };

}

#endif
