/** \file SpectrumView.h
    \brief Declaration of SpectrumView class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_SpectrumView_h
#define profit_SpectrumView_h

#include <cstddef>
#include <functional>
#include <list>
#include <string>
#include <vector>

#include "TGLayout.h"

#include "profit/Energy.h"
#include "profit/Spectrum.h"

#include "RQ_OBJECT.h"
#include "Rtypes.h"

class TF1;
class TGButton;
class TGCompositeFrame;
class TGFrame;
class TGaxis;
class TH1D;
class TLatex;
class TLine;
class TRootEmbeddedCanvas;

namespace profit {

  class OStream;

  /** \class SpectrumView
      \brief Class encapsulating a graphical representation of a single spectrum.
  */
  class SpectrumView {
      RQ_OBJECT("profit::SpectrumView")
      ClassDef(SpectrumView, 0)

    public:
      typedef size_t size_type;

      /** \class Label
          \brief Vertical label appearing on a SpectrumView.
      */
      class Label {
        public:
          /** \brief Create a label associated with an energy.
              \param ienergy The energy.
              \param text The text to display in the label.
          */
          Label(const IEnergy & ienergy, const std::string & text);

          /** \brief Copy-construct a label.
              \param label The source label object to copy.
          */
          Label(const Label & label);

          /// \brief Destroy the label, removing it from the display.
          virtual ~Label();

          /** \brief Display the label at the given coordinates in the currently selected canvas.
              \param x The abscissa at which to display the label, in natural units of the canvas X axis.
              \param y The ordinate at which to display the label, in natural units of the canvas Y axis.
          */
          void display(double x, double y);

          /// \brief Return the energy with which this label is associated.
          const IEnergy & getEnergy() const;

        private:
          Energy m_energy;
          std::string m_text;
          TLatex * m_latex;
          TLine * m_line;
      };

      /// \brief Typedef for container of labels.
      typedef std::list<Label> label_cont_type;

      /// \brief Destroy SpectrumView object.
      virtual ~SpectrumView();

      /// \brief Create an exact duplicate of this object.
      virtual SpectrumView * clone() const = 0;

      /** \brief Return the energy associated with the given domain index in the view.
          \param idx The index.
      */
      virtual Energy getEnergy(size_type idx) const = 0;

      /// \brief Return the mininum energy of the currently viewed (selected) domain.
      virtual Energy getMinEnergy() const = 0;

      /// \brief Return the maxinum energy of the currently viewed (selected) domain.
      virtual Energy getMaxEnergy() const = 0;

      /** \brief Get the abscissa of the given energy in this view.
          \param energy The energy whose abscissa to look up.
      */
      virtual double getAbscissa(const IEnergy & energy) const = 0;

      /** \brief Find positions of peak candidates in energy space.
          \param peak_candidate Output container of peak candidates.
      */
      // Fully qualify profit::Energy because Cint seems to need it.
      virtual void findPeakCandidates(std::vector<profit::Energy> & peak_candidate) const = 0;

      /** \brief Select the energy range of the currently viewed domain. This selection
          does not affect the underlying spectrum, that is, the area excluded by the selection
          is not visible, but it is not discarded.

          If this view's domain is not in energy space, the min/max will be transformed to the
          domain of this view first.
          \param min_energy The minimum energy in the range.
          \param max_energy The maximum energy in the range.
      */
      virtual void select(const IEnergy & min_energy, const IEnergy & max_energy) = 0;

      /// \brief Return number of bins in current selection.
      size_type numSelectedBins() const;

      /** \brief Create and return a new SpectrumView which is the same as this view, except
          that it's domain is the same as this view's currently viewed (selected) domain.

          If this view's domain is not in energy space, the min/max will be transformed to the
          domain of this view first. The caller is responsible for freeing the returned slice.
          \param min_energy The minimum energy in the range.
          \param max_energy The maximum energy in the range.
      */
      virtual SpectrumView * slice(const IEnergy & min_energy, const IEnergy & max_energy) const;

// TODO Fix comments in this and first slice().
      /** \brief Create and return a new SpectrumView which is the same as this view, except
          that its domain is the same as this view's currently viewed (selected) domain.

          If this view's domain is not in energy space, the min/max will be transformed to the
          domain of this view first. The caller is responsible for freeing the returned slice.
          \param min_energy The minimum energy in the range.
          \param max_energy The maximum energy in the range.
      */
      virtual SpectrumView * slice(double min_domain, double max_domain) const;

      /// \brief Return the currently viewed (selected) domain, in natural units for this view type.
      virtual void getDomain(std::vector<double> & domain) const;

      /// \brief Return the currently viewed (selected) range, in natural units for this view type.
      virtual void getRange(std::vector<double> & range) const;

      /// Return the underlying spectrum to which this view refers.
      virtual const Spectrum & getSpectrum() const;

      /** \brief Display this view in the given frame.
          \param parent The frame in which to display the view.
          \param hints Hints for how to justify/expand the frame.
          \param width The width of the view display.
          \param height The height of the view display.
      */
      virtual void display(TGFrame * parent, int hints = kLHintsNormal, long width = 215, long height = 130,
        unsigned int pad_id = 0);

      /// \brief Update the display to reflect all cosmetic changes to the histogram and/or canvas.
      virtual void updateDisplay();

      /// \brief Return true if this view is currently displayed, false otherwise.
      virtual bool isDisplayed() const;

      /// \brief Perform fit to spectrum.
      virtual bool fit();

      /** \brief Find positions of peak candidates in natural units.
          \param peak_candidate Output container of peak candidates.
      */
      virtual void findPeakCandidates(std::vector<double> & peak_candidate) const;

      /** \brief Output fit parameters.
          \param os The stream to which to write the parameters.
      */
      virtual void fitSummary(OStream & os);

      /** \brief Add a textual label to the display, at the position of the given energy.
          \param ienergy The energy at which to locate the label.
          \param text The text to display.
      */
      virtual void addLabel(const IEnergy & ienergy, const std::string & text);

      /// \brief Clear textual labels from the display.
      virtual void clearLabels();

      virtual void applySrcRedshift(double src_redshift);

      virtual double getSrcRedshift() const { return m_src_redshift; }

      virtual bool getHighlight() const;

      virtual void setHighlight(bool highlight = true);

      virtual void scroll(double delta_x);

      virtual bool isVelocity() const = 0;

      virtual void handleClose();

      virtual void handleSelect();

      virtual void smooth();

      static const double s_c;
      static const double s_fwhm_factor;

    protected:
      /** \brief Create SpectrumView object associated with the given spectrum.
          \param spectrum The spectrum.
      */
      SpectrumView(const Spectrum & spectrum, double src_redshift = 0.);

      /** \brief Copy-construct SpectrumView object. Does not display the copied object.
          \param view The SpectrumView object to copy.
      */
      SpectrumView(const SpectrumView & view);

      virtual void init() = 0;

      /** \brief Replace the current member histogram a new histogram with the given domain and range.
          \param domain The new domain.
          \param range The new range.
      */
      virtual void createHistogram(const std::vector<double> & domain, const std::vector<double> & range);

      /// \brief Remove this view from the display to which it is attached.
      virtual void clearDisplay();

      /** \brief Select the currently viewed domain, in natural units.
          \param min The minimum domain value.
          \param max The maximum domain value.
      */
      virtual void selectDomain(double min, double max);

      /// \brief Display all labels currently owned by this view on the view canvas.
      virtual void displayLabels();

      label_cont_type m_label;
      std::vector<double> m_fit_par;
      std::vector<double> m_fit_par_error;
      std::string m_canvas_name;
      std::string m_hist_name;
      std::string m_func_name;
      double m_src_redshift;
      int m_hints;
      long m_width;
      long m_height;
      unsigned int m_pad_id;
      const Spectrum * m_spectrum;
      TH1D * m_hist;
      TGaxis * m_top_axis;
      TF1 * m_fit_func;
      TGLayoutHints * m_layout_hints;
      TGCompositeFrame * m_main_frame;
      TGCompositeFrame * m_button_frame;
      TGButton * m_close_button;
      TGButton * m_select_button;
      TRootEmbeddedCanvas * m_canvas;
      bool m_own_canvas;
      bool m_highlight;
      bool m_displayed;
  };

  /** \class EnergyView
      \brief Class encapsulating a graphical representation of a single spectrum in energy space.
  */
  class EnergyView : public SpectrumView {
    public:
      /** \brief Create a EnergyView of the given spectrum.
          \param spectrum The spectrum being viewed.
      */
      EnergyView(const Spectrum & spectrum, double src_redshift = 0.);

      /** \brief Create a EnergyView from a generic SpectrumView.
          \param view The view being copied.
      */
      EnergyView(const SpectrumView & view);

      /// \brief Create an exact duplicate of this object.
      virtual EnergyView * clone() const;

      /** \brief Return the energy associated with the given domain index in the view.
          \param idx The index.
      */
      virtual Energy getEnergy(size_type idx) const;

      /// \brief Return the mininum energy of the currently viewed (selected) domain.
      virtual Energy getMinEnergy() const;

      /// \brief Return the maxinum energy of the currently viewed (selected) domain.
      virtual Energy getMaxEnergy() const;

      /** \brief Get the abscissa of the given energy in this view.
          \param energy The energy whose abscissa to look up.
      */
      virtual double getAbscissa(const IEnergy & energy) const;

      /** \brief Find positions of peak candidates in energy space.
          \param peak_candidate Output container of peak candidates.
      */
      virtual void findPeakCandidates(std::vector<Energy> & peak_candidate) const;

      /** \brief Select the energy range of the currently viewed domain. This selection
          does not affect the underlying spectrum, that is, the area excluded by the selection
          is not visible, but it is not discarded.

          If this view's domain is not in energy space, the min/max will be transformed to the
          domain of this view first.
          \param min_energy The minimum energy in the range.
          \param max_energy The maximum energy in the range.
      */
      virtual void select(const IEnergy & min_energy, const IEnergy & max_energy);

      /** \brief Find positions of peak candidates in natural units.
          \param peak_candidate Output container of peak candidates.
      */
      virtual void findPeakCandidates(std::vector<double> & peak_candidate) const;

      virtual bool isVelocity() const;

    protected:
      virtual void init();

      /** \brief Replace the current member histogram a new histogram with the given domain and range.
          \param domain The new domain.
          \param range The new range.
      */
      virtual void createHistogram(const std::vector<double> & domain, const std::vector<double> & range);
  };

  /** \class WavelengthView
      \brief Class encapsulating a graphical representation of a single spectrum in energy space.
  */
  class WavelengthView : public SpectrumView {
    public:
      /** \brief Create a WavelengthView of the given spectrum.
          \param spectrum The spectrum being viewed.
      */
      WavelengthView(const Spectrum & spectrum, double src_redshift = 0.);

      /** \brief Create a WavelengthView from a generic SpectrumView.
          \param view The view being copied.
      */
      WavelengthView(const SpectrumView & view);

      /// \brief Create an exact duplicate of this object.
      virtual WavelengthView * clone() const;

      /** \brief Return the energy associated with the given domain index in the view.
          \param idx The index.
      */
      virtual Energy getEnergy(size_type idx) const;

      /// \brief Return the mininum energy of the currently viewed (selected) domain.
      virtual Energy getMinEnergy() const;

      /// \brief Return the maxinum energy of the currently viewed (selected) domain.
      virtual Energy getMaxEnergy() const;

      /** \brief Get the abscissa of the given energy in this view.
          \param energy The energy whose abscissa to look up.
      */
      virtual double getAbscissa(const IEnergy & energy) const;

      /** \brief Find positions of peak candidates in energy space.
          \param peak_candidate Output container of peak candidates.
      */
      virtual void findPeakCandidates(std::vector<Energy> & peak_candidate) const;

      /** \brief Select the energy range of the currently viewed domain. This selection
          does not affect the underlying spectrum, that is, the area excluded by the selection
          is not visible, but it is not discarded.

          If this view's domain is not in energy space, the min/max will be transformed to the
          domain of this view first.
          \param min_energy The minimum energy in the range.
          \param max_energy The maximum energy in the range.
      */
      virtual void select(const IEnergy & min_energy, const IEnergy & max_energy);

      /** \brief Find positions of peak candidates in natural units.
          \param peak_candidate Output container of peak candidates.
      */
      virtual void findPeakCandidates(std::vector<double> & peak_candidate) const;

      virtual bool isVelocity() const;

    protected:
      virtual void init();

      /** \brief Replace the current member histogram a new histogram with the given domain and range.
          \param domain The new domain.
          \param range The new range.
      */
      virtual void createHistogram(const std::vector<double> & domain, const std::vector<double> & range);
  };

  /** \class VelocityView
      \brief Class encapsulating a graphical representation of a single spectrum in velocity space.
  */
  class VelocityView : public SpectrumView {
    public:
      // Unary functor to convert energy to velocity space.
      class EnergyToVelocity : std::unary_function<double, double> {
        public:
          EnergyToVelocity(const IEnergy & origin): m_lambda_0(origin.wavelength()) {}

          double operator ()(double energy) {
            double lambda = Energy(energy).wavelength(); // Note: this never returns 0, nor blows up for energy = 0.
            return (lambda - m_lambda_0) / lambda * s_c;
          }

        private:
          double m_lambda_0;
      };

      // Unary functor to convert velocity to energy space.
      class VelocityToEnergy : std::unary_function<double, double> {
        public:
          VelocityToEnergy(const IEnergy & origin): m_energy_0(origin.energy()) {}

          double operator ()(double velocity) {
            return m_energy_0 * (1. - velocity / s_c);
          }

        private:
          double m_energy_0;
      };

      /** \brief Create a VelocityView of the given spectrum.
          \param spectrum The spectrum being viewed.
          \param origin The energy used as the origin in velocity space. v(lambda) = (lambda - origin.lambda()) / lambda.
      */
      VelocityView(const Spectrum & spectrum, const IEnergy & origin, double src_redshift = 0.);

      /** \brief Create a VelocityView of the given spectrum.
          \param view Another view of the spectrum.
          \param origin The energy used as the origin in velocity space. v(lambda) = (lambda - origin.lambda()) / lambda.
      */
      VelocityView(const SpectrumView & view, const IEnergy & origin);

      /// \brief Create an exact duplicate of this object.
      virtual VelocityView * clone() const;

      /** \brief Return the energy associated with the given domain index in the view.
          \param idx The index.
      */
      virtual Energy getEnergy(size_type idx) const;

      /// \brief Return the mininum energy of the currently viewed (selected) domain.
      virtual Energy getMinEnergy() const;

      /// \brief Return the maxinum energy of the currently viewed (selected) domain.
      virtual Energy getMaxEnergy() const;

      /** \brief Get the abscissa of the given energy in this view.
          \param energy The energy whose abscissa to look up.
      */
      virtual double getAbscissa(const IEnergy & energy) const;

      /** \brief Find positions of peak candidates in energy space.
          \param peak_candidate Output container of peak candidates.
      */
      virtual void findPeakCandidates(std::vector<Energy> & peak_candidate) const;

      /// \brief Return the central wavelength which was used to produce this view.
      virtual const Wavelength & getOrigin() const;

      /** \brief Select the energy range of the currently viewed domain. This selection
          does not affect the underlying spectrum, that is, the area excluded by the selection
          is not visible, but it is not discarded.

          If this view's domain is not in energy space, the min/max will be transformed to the
          domain of this view first.
          \param min_energy The minimum energy in the range.
          \param max_energy The maximum energy in the range.
      */
      virtual void select(const IEnergy & min_energy, const IEnergy & max_energy);

      /** \brief Find positions of peak candidates in natural units.
          \param peak_candidate Output container of peak candidates.
      */
      virtual void findPeakCandidates(std::vector<double> & peak_candidate) const;

      virtual bool isVelocity() const;

    protected:
      virtual void init();

      /** \brief Replace the current member histogram a new histogram with the given domain and range.
          \param domain The new domain.
          \param range The new range.
      */
      virtual void createHistogram(const std::vector<double> & domain, const std::vector<double> & range);

      Wavelength m_origin;
  };

}

#endif
