/** \file Table.h
    \brief Declaration of Table class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_Table_h
#define profit_Table_h

#include <cstdio>
#include <string>
#include <vector>

namespace profit {

  /** \class TableFilePtr
      \brief Completely internal data file table accessor.
  */
  class TableFilePtr;

  /** \class Record
      \brief Class encapsulating the idea of a single record in a table.
  */
  class Record {
    public:
      typedef size_t size_type;

      /** \brief Create a Record associated with the given table data file.
          \param fp The table data file pointer.
          \param record_index The record identifier in the table.
      */
      Record(const TableFilePtr * fp, size_type record_index);

      /** \brief Return the contents of the given field in this record. If the field
          cannot be converted to a double, this throws an exception.
          \param field_name The name of the field.
      */
      double getField(const std::string & field_name) const;

      /** \brief Return the contents of the given field in this record as a string.
          \param field_name The name of the field.
          \param field_value The value of the field.
      */
      void getField(const std::string & field_name, std::string & field_value) const;

      /** \brief Return the contents of the given field in this record as a vector.
          \param field_name The name of the field.
          \param field_value The value of the field.
      */
      void getField(const std::string & field_name, std::vector<double> & field_value) const;

    private:
      const TableFilePtr * m_fp;
      size_type m_index;
  };

  /** \class InputTable
      \brief Class encapsulating read-only access to data tables.
  */
  class InputTable {
    public:
      typedef Record::size_type size_type;

      /** \brief Create an file accessor object.
          \param file_name The name of the file.
          \param table_name The name of the table in the file.
      */
      InputTable(const std::string & file_name, const std::string & table_name);

      ~InputTable();

      /// \brief Return the number of records in the table.
      size_type getNumRecords() const;

      /** \brief Get the contents of a table column. Column values must all be convertible to double.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      void getColumn(const std::string & field_name, std::vector<double> & column) const;

      /** \brief Get the contents of a table column as strings.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      void getColumn(const std::string & field_name, std::vector<std::string> & column) const;

      /** \brief Get the dimensionality of the given field.
          \param field_name The name of the field/column.
      */
      size_type getFieldDim(const std::string & field_name) const;

      /** \brief Return the content of a keyword, as a string.
          \param key_name The name of the keyword to get.
      */
      std::string getKeyword(const std::string & key_name) const;

      /// \brief Return the name of the file containing this table.
      const std::string & getFileName() const;

      /// \brief Return the name of the directory containing this table.
      std::string getDirectoryName() const;

      /** \brief Return a single record (row) of the file.
          \param record_index The index (0-based row number) in the file.
      */
      Record getRecord(size_type record_index) const;
     
      /** \brief Filter the current table contents using the given expression.
          \param expression The expression (Cfitsio row filtering syntax.)
      */
      void filter(const std::string & expression);

      /** \brief Filter the current table contents using the given expression.
          \param field_name The name of the field on which to filter.
          \param min_value The minimum value of the filter range.
          \param max_value The maximum value of the filter range.
      */
      void filter(const std::string & field_name, double min_value, double max_value);

    private:
      TableFilePtr * m_fp;
  };

}

#endif

