/** \file UnitTest.h
    \brief Declaration of UnitTest class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_UnitTest_h
#define profit_UnitTest_h

#include <iostream>
#include <string>

#include "profit/Application.h"

namespace profit {

  class AtomicLevel;
  class AtomicTransition;
  class IEnergy;
  class Spectrum;

  /** \class UnitTest
      \brief Encapsulation of the unit test for profit.
  */
  class UnitTest : public Application {
    public:
      /// \brief Create a unit test.
      UnitTest();

      /** \brief Destroy this unit test. */
      virtual ~UnitTest();

      /// \brief Perform tests of the Table class, for file i/o.
      void testTable();

      /// \brief Perform tests of Energy-related classes.
      void testEnergy();

      /// \brief Perform tests of the AtomDb class.
      void testAtomDb();

      /// \brief Perform tests of the PlasmaDb class.
      void testPlasmaDb();

      /// \brief Perform tests of the Spectrum class.
      void testSpectrum();

      /// \brief Perform tests of the SpecFile class.
      void testSpecFile();

      /// \brief Perform tests of the SpectrumView class.
      virtual void testSpectrumView();

      /// \brief Perform tests of the Gui class.
      virtual void testGui();

      /// \brief Perform tests of stream-related classes.
      virtual void testStream();

      /** \brief Run a single test correctly depending on whether debug mode is set.
          \param test The test to run.
      */
      void runTest(void (UnitTest::* test)());

    protected:
      /// \brief Perform the unit test.
      virtual void runApp();

      /** \brief Check whether a file exists.
          \param file_name the file to check.
      */
      bool fileExists(const std::string & file_name) const;

      /** \brief Determine whether two values are roughly equal. Return true if they are equal,
          false otherwise.
          \param val1 The first value to compare.
          \param val2 The second value to compare.
          \param abs_epsilon The absolute tolerance used to determine equality.
          \param rel_epsilon The relative, or fractional tolerance used to determine equality.
      */
      bool isEqual(double val1, double val2, double abs_epsilon = 0., double rel_epsilon = 0.) const;

      /** \brief Determine whether two values are equal within an absolute tolerance of +/-epsilon.
          Return true if they are equal, false otherwise.

          This returns the truth of the expression epsilon > |val1 - val2|
          \param val1 The first value to compare.
          \param val2 The second value to compare.
          \param epsilon The absolute tolerance used to determine equality.
      */
      bool isAbsEqual(double val1, double val2, double epsilon = 0.) const;

      /** \brief Determine whether two values are equal within an relative tolerance of +epsilon.
          Return true if they are equal, false otherwise.

          This returns the truth of the expression 1. < (val1 / val2) < 1. + epsilon if |val1| > |val2|, or
          1. < (val2 / val1) < 1. + epsilon if |val2| > |val1|.
          \param val1 The first value to compare.
          \param val2 The second value to compare.
          \param epsilon The relative, or fractional tolerance used to determine equality.
      */
      bool isRelEqual(double val1, double val2, double epsilon = 0.) const;

      /// \brief Set up a "hit" on this process, to kill the main application in case it hangs.
      bool hitMe();

      /** \brief Determine whether an energy object constructed from a given set of values behaves properly.
          \param correct_energy The value which should be returned by the energy method.
          \param correct_wavelength The value which should be returned by the wavelength method.
      */
      void testEnergy(double correct_energy, double correct_wavelength);

      /** \brief Determine whether an energy object has the correct wavelength and frequency.
          \param hint Hint about the conditions of the test, used to report context-sensitive errors.
          \param ienergy The energy object.
          \param correct_energy The value which should be returned by the energy method.
          \param correct_wavelength The value which should be returned by the wavelength method.
      */
      void testEnergy(const std::string & hint, const IEnergy & ienergy, double correct_energy, double correct_wavelength) const;

      /** \brief Determine whether an AtomicLevel object has the correct values for its members.
          \param hint Hint about the conditions of the test, used to report context-sensitive errors.
          \param level The AtomicLevel object being checked.
          \param energy The correct energy.
          \param quantumN The correct principle quantum number.
          \param quantumL The correct orbital angular momentum quantum number.
          \param quantumS The correct spin angular momentum quantum number.
          \param degeneracy The correct degeneracy for this state.
          \param config String describing the configuration.
      */
      void testAtomicLevel(const std::string & hint, const AtomicLevel & level, double energy, int quantumN, int quantumL,
        double quantumS, int degeneracy, const std::string & config) const;

      /** \brief Determine whether an AtomicTranstion object has the correct levels and the correct
          populations of collisional, radiative and satellite transitions.
          \param hint Hint about the conditions of the test, used to report context-sensitive errors.
          \param transition The transition being tested.
          \param lower_level The correct lower level involved in the transtion.
          \param upper_level The correct upper level involved in the transtion.
          \param num_coll The correct number of collisional processes.
          \param num_rad The correct number of radiative processes.
          \param num_sat The correct number of satellite processes.
      */
      void testAtomicTransition(const std::string & hint, const AtomicTransition & transition,
        const AtomicLevel & lower_level, const AtomicLevel & upper_level, size_t num_coll, size_t num_rad, size_t num_sat) const;

      /** \brief Determine whether a Spectrum object has the correct state.
          \param hint Hint about the conditions of the test, used to report context-sensitive errors.
          \param spectrum The Spectrum object being tested.
          \param num_elements The correct size for the Spectrum's energy containers.
      */
      void testSpectrum(const std::string & hint, const Spectrum & spectrum, size_t num_elements) const;

      void testTextFile(const std::string & text_file_name) const;

      static double s_double_epsilon;
      static double s_float_epsilon;
      std::string m_atom_db_env;
      std::string m_atom_db_file_name;
      std::string m_xstar_db_file_name;
      std::string m_spectrum_file_name;
      std::string m_ebounds_file_name;
      std::string m_text_spectrum_file_name;
      std::string m_unfmt_file_name;
      std::string m_unordered_text_file_name;
  };

}

#endif
