/** \file UnitTestGui.h
    \brief Declaration of UnitTestGui class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_UnitTestGui_h
#define profit_UnitTestGui_h

#include <list>
#include <string>
#include <vector>

#include "RQ_OBJECT.h"
#include "Rtypes.h"
#include "TQObject.h"

#include "profit/UnitTest.h"

class TGFrame;
class TGMainFrame;

namespace profit {

  class Gui;
  class SpectrumView;

  /** \class UnitTestGui
      \brief Encapsulation of the unit test for profit.
  */
  class UnitTestGui : public UnitTest {
    public:
      RQ_OBJECT("profit::UnitTestGui")
      ClassDef(UnitTestGui, 0)

      /// \brief Create a graphical unit test.
      UnitTestGui();

      /// \brief Destroy the graphical unit test.
      virtual ~UnitTestGui();

      /// \brief Perform the actual unit test.
      virtual void runAllTests();

      /// \brief Perform tests of the SpectrumView class.
      virtual void testSpectrumView();

      /// \brief Perform tests of the Gui class.
      virtual void testGui();

      /// \brief Perform tests of stream-related classes.
      virtual void testStream();

    private:
      typedef std::list<TGFrame *> FrameCont;

      /// \brief Display the GUI which performs the test.
      virtual void runApp();

      /** \brief Determine whether a SpectrumView object has the correct state.
          \param hint Hint about the conditions of the test, used to report context-sensitive errors.
          \param spectrum_view The SpectrumView object being tested.
          \param domain Correct domain container.
          \param range Correct range container.
      */
      void testSpectrumView(const std::string & hint, const SpectrumView & spectrum_view, const std::vector<double> & domain,
        const std::vector<double> & range) const;

      /** \brief Add a frame to the container of all frames. Throw exception if frame pointer is 0,
          in which case it is assumed dynamic allocation failed.
          \param context A hint to be use in exception.
          \param frame The frame to add.
      */
      void storeFrame(const std::string & context, TGFrame * frame);

      /** \brief Add a frame to a frame container. Throw exception if frame pointer is 0,
          in which case it is assumed dynamic allocation failed.
          \param context A hint to be use in exception.
          \param frame The frame to add.
          \param cont The container to free.
      */
      void storeFrame(const std::string & context, TGFrame * frame, FrameCont & cont);

      /// \brief Free up all resources dynamically allocated by the test.
      void clear();

      /** \brief Free up frames within a container.
          \param cont The container.
      */
      void clearFrames(FrameCont & cont);

      FrameCont m_frame_cont;
      FrameCont m_temp_frame_cont;
      TGMainFrame * m_main_frame;
      SpectrumView * m_spectrum_view;
      Gui * m_gui;
      bool m_test_ran;
  };

}

#endif
